import sys
PY2 = sys.version_info[0] == 2
if not PY2:
    unicode = str

class ModelCommonParent():

    @classmethod
    def orderAll(cls,q):
        return q

    @classmethod
    def all(cls,DBSession,orderby=None,orderdir='asc'):
        q = cls.orderAll(DBSession.query(cls))
        if orderby is not None:
            if orderdir == 'asc':
                q = q.order_by(getattr(cls,orderby).asc() )
            else:
                q = q.order_by((getattr(cls,orderby).desc()  ))
        return q.all()
        #return cls.orderAll(DBSession.query(cls)).all()
    
    @classmethod
    def byId(cls,id,DBSession):
        return DBSession.query(cls).filter(cls.id==id).first()

    @classmethod
    def byField(cls,field,value,DBSession,orderby=None,orderdir='asc'):
        q = DBSession.query(cls).filter(getattr(cls, field)==value)
        if orderby is not None:
            if orderdir == 'asc':
                q = q.order_by(getattr(cls,orderby).asc() )
            else:
                q = q.order_by((getattr(cls,orderby).desc()  ))
        return q.all()
    
    @classmethod
    def byFields(cls,fields,DBSession,orderby=None,orderdir='asc'):
        q = DBSession.query(cls)
        for i,field in enumerate(fields):
            q = q.filter(getattr(cls, field[0])==field[1])
        if orderby is not None:
            if orderdir == 'asc':
                q = q.order_by(getattr(cls,orderby).asc() )
            else:
                q = q.order_by((getattr(cls,orderby).desc()  ))
        return q.all()


    @classmethod
    def delete(cls,element,DBSession):
        DBSession.delete(element)
    @classmethod
    def deleteAll(cls,elements,DBSession):
        for e in elements:
            DBSession.delete(e)


    def __str__(self):
        return "<{}({})>".format(self.__class__.__name__,self.id) #unicode(self)

    def __repr__(self):
        return unicode(self)

    def __unicode__(self):
        return unicode(str(self))



from sqlalchemy import engine_from_config
from sqlalchemy.orm import sessionmaker
import configparser,os

import logging,logging.config
l = logging.getLogger(__name__)

def engineFromConfigPath(config_uri):

  config = configparser.ConfigParser() 
  with open(config_uri,"r") as inifile:
    config.read_string(inifile.read())
  ##TODO: here is wrong. what should it be?
  here = os.getcwd()
  config["app:main"]["here"] = here
  logging.config.fileConfig(config)
  return config,engineFromConfig(config)

def engineFromConfig(config):
  engine = engine_from_config(config["app:main"], "sqlalchemy.")
  return engine

def engineFromSettings(config):
  engine = engine_from_config(config, "sqlalchemy.")
  return engine

def factoryFormSettings(config):
  engine = engineFromSettings(config)
  session_factory = sessionmaker(bind=engine)
  return session_factory


import csv
import codecs
try:
    from StringIO import StringIO
except Exception as e:
    from io import StringIO
import tempfile


from six import PY2, iteritems
if not PY2:
    unicode = str
    # from six import u as unicode

import logging,datetime
l = logging.getLogger("ppss.pyramidutils.utf8csv")


class UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode("utf-8")

class UnicodeReader:
    def __init__(self, f, dialect=csv.excel, encoding="utf-8-sig", **kwds):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, dialect=dialect, **kwds)

    def next(self):
        '''next() -> unicode
        This function reads and returns the next line as a Unicode string.
        '''
        row = self.reader.next()
        return [unicode(s, "utf-8") for s in row]

    def __iter__(self):
        return self


class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()
        self.l = logging.getLogger(self.__class__.__name__)
        #self.l = logging.getLogger("main")

    def writerow(self, row):
        try:
            stringrow = list(map(lambda s: str(s) if isinstance(s, (int, float) ) else s, row ))
            r = 1
            self.writer.writerow([s.encode("utf-8") for s in stringrow])
            # Fetch UTF-8 output from the queue ...
            r = 2
            data = self.queue.getvalue()
            r = 3
            data = data.decode("utf-8")
            # ... and reencode it into the target encoding
            r = 4
            data = self.encoder.encode(data)
            # write to the target stream
            r = 5
            self.stream.write(data)
            # empty queue
            r = 6
            self.queue.truncate(0)
        except Exception as e:
            #print row,e,r
            #raise e
            self.l.error("inner errore, errore!!")
            self.l.exception("step:"+unicode(r)+ " " + e.message + repr(e)+ "--" + repr(row)    )
            pass

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)

class Exporter(object):
    def __init__(self,evcollection,titles,parserows=True,delimiter = ',',getter=None,datetimeformat = "%Y-%m-%d"):
        #self.evl = [x.getMap() for x in evcollection]
        self.titles = titles

        if parserows:
            evrows = []
            for ev in evcollection:
                row = []
                for k in self.titles:
                    try:
                        if getter:
                            val = getter(ev,k)
                        else:
                            val = ev.getVal(k)
                    except Exception as e:
                        l.warning("key {key} not in row {row}, exception {e}".format(key=k, row=str(row), e=e ))
                        val = ""
                    if isinstance(val,datetime.datetime) or isinstance(val,datetime.date):
                        val = unicode(val.strftime(datetimeformat))
                    if not (isinstance(val,str) or isinstance(val,unicode)):
                        val = unicode(val)
                    row.append(val)
                evrows.append(row)
        else:
            evrows = evcollection
        self.evl = evrows
        self.delimiter = delimiter
        if PY2:
            self.retfile = tempfile.NamedTemporaryFile()
        else:
            self.retfile = tempfile.NamedTemporaryFile(mode='w',encoding="utf8")


    def writeAll(self, delimiter = None):
        d = delimiter if delimiter else self.delimiter if self.delimiter else ','
        if PY2:
            w = UnicodeWriter(self.retfile, delimiter=d, quoting=csv.QUOTE_ALL)
        else:
            w = csv.writer(self.retfile, dialect=csv.excel, delimiter = d)
        w.writerow(self.titles)
        for maprow in self.evl:
            l.info("*******{row}".format(row=maprow) )
            w.writerow(maprow)
        self.retfile.seek(0, 0)
        return self.retfile


class Importer(object):
    titlepos = {}
    rows = []
    header = []

    def __init__(self, fn, mapping=None, delimiter=",", headertransform=None,header=[]):
        self.titlepos = {}
        self.rows = []
        #f = codecs.open(fn, "r", "utf-8")  #open(fn,'r')
        if PY2:
            self.csvreader = UnicodeReader(open(fn,'rb'),delimiter=delimiter)
        else:
            self.csvreader = csv.reader(open(fn,'r',encoding='utf-8'), dialect=csv.excel,delimiter=delimiter)
        #self.csvreader = csv.reader(f,delimiter=";")
        if header:
            self.header = header
        else:
            h = next(self.csvreader)
            if headertransform:
                h = [headertransform(x) for x in h]
            self.header=h 
        l.debug("header:"+repr(self.header))
        for i in range(len(self.header)):
            self.titlepos[ self.header[i].strip() ] = i
        for row in self.csvreader:
            try:
                thisrow = {}
                for key,val in iteritems(self.titlepos):
                    if mapping :
                        l.debug(mapping)
                        thisrow[mapping[key]] = row[val].strip()
                    else:
                        l.debug(key)
                        thisrow[key] = row[val].strip()

                l.debug("*read:"+repr(thisrow))
                self.rows.append(thisrow)
            except Exception as  e:
                l.error("error reading:{error}\nrow:{row}".format(error=repr(e),row=row) )
                raise e


    def getRows(self):
        return self.rows




from .utils import SettingsReader
import logging
l = logging.getLogger("ppss.pyramidutils")
import os,shutil,uuid,re,string,unicodedata

from six import PY2
if not PY2:
    from six import u as unicode

_valid_filename_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)

class FileManager(SettingsReader):
    savepath = "/tmp/finale"
    tmppath = "/tmp"
    myconf = ['savepath','tmppath']

    def __init__(self):
        raise Exception("can't instantiate FileManager class. Only class methods available")

    @classmethod
    def confended(cls):
        os.makedirs(cls.tmppath,exist_ok = True)
        os.makedirs(cls.savepath,exist_ok = True)

    @classmethod
    def sanitizeFilename(cls,filename, whitelist=_valid_filename_chars, replace=' '):
        """Better safe than sorry: sanitize a filename replacing unsafe chars. 
            @filename: string to sanitize
            @whitelist: string containing safe chars in filename
            @replace: string containg valid chars that we prefer to substitute with an _
            returns: sanitized string"""
        # replace spaces
        for r in replace:
            filename = filename.replace(r,'_')
        
        # keep only valid ascii chars
        cleaned_filename = unicodedata.normalize('NFKD', unicode(filename)).encode('ASCII', 'ignore').decode()
        
        # keep only whitelisted chars
        return ''.join(c for c in cleaned_filename if c in whitelist)

    @classmethod
    def slugify(cls,filename):
        """
        Normalizes string, converts to lowercase, removes non-alpha characters,
        and converts spaces to hyphens.
        """
        import unicodedata
        value = unicodedata.normalize('NFKD', unicode(filename)).encode('ascii', 'ignore')
        value = unicode(re.sub('[^\w\s-]', '', value).strip().lower())
        value = unicode(re.sub('[-\s]+', '-', value))
        return value

    @classmethod
    def saveToTmp(cls,requestfile):
        infile = requestfile.file
        #make sure it's not a malicious attack
        name = FileManager.sanitizeFilename( str(requestfile.name) )

        file_path = os.path.join(cls.tmppath, str(uuid.uuid4()) + name )
        l.debug("FileManager.saveToTmp path={path}".format(path=file_path))
        temp_file_path = file_path + '~'
        if not os.path.exists(os.path.dirname(temp_file_path)):
            os.makedirs(os.path.dirname(temp_file_path))
        output_file = open(temp_file_path, 'wb')

        infile.seek(0)
        while True:
            data = infile.read(2<<16)
            if not data:
                break
            output_file.write(data)
        output_file.close()
        os.rename(temp_file_path, file_path)
        return file_path

    @classmethod
    def moveToDestination(cls,source,filename,subfolder=""):
        target = os.path.join(cls.savepath,subfolder,filename)
        l.debug("target filename:{target}".format(target=target))
        if not os.path.exists(os.path.dirname(target)):
            os.makedirs(os.path.dirname(target))

        os.rename(source,target)
        return target

    @classmethod
    def deleteFile(cls,filename):
        os.remove(filename)




import logging
l = logging.getLogger(__name__)

class jsonwalker:
    def __init__(self, callback, skip_keys:list = []) -> None:
        self.callback = callback
        self.skip_keys = set(skip_keys)
        #try:
        #    iterator = iter(callback)
        #    self.callback = callback
        #except TypeError:
        #    self.callback = (callback,)

    def explorelist(self, node):
        newlist = []
        for item in node:
            newlist.append(self.explore(item))
        return newlist

    def exploredict(self, node):
        struct = {}
        for k, v in node.items():
            if k in self.skip_keys:
                struct[k] = v
            else:    
                struct[k] = self.explore(v)
        return struct

    

    def explore(self, node):
        if isinstance(node, dict):
            return self.exploredict(node)
        elif isinstance(node, list):
            return self.explorelist(node)
        else:
            # return self.encryptval(node)
            return self.callback(node)

    def walk(self, thejson):
        return self.explore(thejson)



from sqlalchemy import engine_from_config
from sqlalchemy.exc import DisconnectionError,InvalidRequestError,OperationalError

import time
from datetime import datetime,timedelta
import queue,threading


from .sessionhandling import (engineFromSettings,factoryFormSettings,session_scope)

import logging
l = logging.getLogger(__name__)


#from pyramid.threadlocal import get_current_registry
#from pyramid.request import Request


def threaded_jobs(settings,threadqueue,mynumber,callback):
  factory = factoryFormSettings(settings)
  while True:
    l.debug("executor is looping")
    try:
      job = threadqueue.get(timeout=120.0)
      try:
        with session_scope(factory) as dbsession:
          callback(job,dbsession,factory)
      except (DisconnectionError, InvalidRequestError,OperationalError) as dbconnerror:
        l.exception("db connection died of unexpected violent death. Trying to revive it...")
        if job is not None:
          __incomingjobsqueue.put(job)
          l.info("putting the job back in queue {}".format(job))
        time.sleep(60)
        factory = factoryFormSettings(settings)
        raise dbconnerror
    except queue.Empty as nojob:
      l.debug("no job received")
    except Exception as e:
      l.exception("Exception, but still runing!")




__queuelock = threading.Semaphore()
__allqueues = {}
__defaultqueue = None

def getQueue(name=None):
  with __queuelock:
    if name is None:
      return __defaultqueue
    else:
      return __allqueues.get(name)

def __addQueue(queue,name):
  with __queuelock:
    global __defaultqueue
    if __defaultqueue is None:
      __defaultqueue = queue
    if name in __allqueues:
      raise Exception("Name {name} already assigned in queues".format(name=name))
    __allqueues[name] = queue


def startThread(settings,callback,name="executor",concurrent=1):
  threadqueue = queue.Queue()
  __addQueue  (threadqueue,name)
  for i in range(concurrent):
    x = threading.Thread(name=name,target=threaded_jobs, args=( settings,threadqueue,i,callback ) )
    x.setDaemon(True)
    x.start()
  return threadqueue


def threaded_job_loop(settings,callback,interval,fromstart,canoverlap):
  factory = factoryFormSettings(settings)
  startdt = datetime.now()
  td = timedelta(seconds = interval)
  while True:
    l.debug("executor is looping")
    try:
      try:
        startdt = datetime.now()
        with session_scope(factory) as dbsession:
          callback(dbsession,factory)  
      except (DisconnectionError, InvalidRequestError,OperationalError) as dbconnerror:
          l.exception("db connection died of unexpected violent death. Trying to revive it...")
          time.sleep(10)
          factory = factoryFormSettings(settings)
      except Exception as e:
        l.exception("Exception, but still runing!")
      try:
        if not fromstart:
          time.sleep(interval)
        else:
          now = datetime.now()
          nextrun = now + td
          if nextrun<now:
            time.sleep( ((nextrun) - now).total_seconds()  )
      except Exception as e:
        l.exception("Exception waiting next run, but still runing!")
        time.sleep(interval)
    except Exception as e:
      l.exception("Exception in outer loop. Still running")
          

def startLoop(settings,callback,name="bgworker",interval=60,fromstart=False,canoverlap=False):
  x = threading.Thread(name=name,target=threaded_job_loop, args=( settings,callback,interval,fromstart,canoverlap ) )
  x.setDaemon(True)
  x.start()





from .utils import SettingsReader
import logging
l = logging.getLogger(__name__)



class AppSettings(SettingsReader):
    myconf = ['adminname','adminpass','customtemplatesrc']
    customtemplatesrc = '/tmp'


__allsettings={}

def initAppSettings(prefix,classname):
  newclass = type(classname  , (SettingsReader,),{})
  __allsettings [classname  ] = newclass 



from .utils import SettingsReader
from .utils import SettingsReader as Utils  ##alias for backward compatibility
from .filemanager import FileManager
from .appsettings import AppSettings
from .modelbase import ModelCommonParent
from .utf8csv import (Importer,Exporter)
from .backgroundjobs import getQueue, startThread
from .sessionhandling import engineFromSettings,factoryFormSettings,session_scope
from .scriptbase import engineFromConfigPath


import logging
import types
#from six import PY2

l = logging.getLogger("ppss.pyramidutils")

import sys
#import types
PY2 = sys.version_info[0] == 2

if PY2:
    def is_nonstr_iter(v):
        return hasattr(v, '__iter__')
else:
    def is_nonstr_iter(v):
        if isinstance(v, str):
            return False
        return hasattr(v, '__iter__')
string = types.StringTypes if PY2 else str


class SettingsReader():
    myconf = []

    @classmethod
    def confended(cls, **kwargs):
        pass

    @classmethod
    def valueread(cls,prop,value):
        return value

    @classmethod
    def __getMyConf(cls,settings, prefix,defaultval=None,**kwargs):
        resvals = {}
        resdicts = {}
        #string = types.StringTypes if PY2 else str
        l.debug("reading conf for {}".format(prefix))
        for k in cls.myconf:
            if isinstance(k, string):
                prop = k
                key = prefix+"."+k
                default = getattr(cls, k, defaultval)
            else:
                try:
                    prop = k[0]
                    key = prefix+"."+k[0]
                    default = k[1]
                except Exception:
                    l.warn("exception reading {}".format(k))
                    continue

            value = settings.get(key, default)
            value = cls.valueread(prop,value)
            if "." in prop:
                propparts = prop.split(".")

                dictname = propparts[0]
                dictkey = propparts[1]
                if dictname not in resdicts:
                    resdicts[dictname] = {}
                resdicts[dictname][dictkey] = value
            else:
                setattr(cls, prop, value)
                if key in settings:
                    l.debug("value of {key} set to: {val}".format(
                        key=prop, val=value))

        for key, value in resdicts.items():
            resvals[key]=value
            setattr(cls, key, value)
            l.debug("value of {key} set to: {val}".format(key=key, val=value))
            #setattr(cls,k,unicode(settings[key]) )
        for k in cls.myconf:
            if isinstance(k, string):
                key = k
            else:
                key = k[0]
            l.debug("val for {key} = {val}".format(
                key=key, val=getattr(cls, key)))
        return resvals

    def __getAllConf(cls,settings ,prefix,defaultval,**kwargs):
        prefixdotted = prefix + "."
        prefixlen = len(prefix) + 1
        parse = kwargs.get('parse',{})
        resdicts = {}
        for k in settings.keys():
            if not k.startswith(prefixdotted):
                continue
            key = k[prefixlen:]
            value = settings.get(k)
            if key in parse:
                value = parse[key](value)
            resdicts[key]=value
            cls.myconf.append(key)
        for key, value in resdicts.items():
            setattr(cls, key, value)
            resvals[key]=value
            l.debug("value of {key} set to: {val}".format(key=key, val=value))
        return resvals

    @staticmethod
    def valueAsList(value,resaslist = False,stripit = True,mapfunc=None):
        if not is_nonstr_iter(value):
            #print(repr(confval))
            value = aslist(value,flatten=False)
        if resaslist != False and isinstance(resaslist, string):
            value = value.split(resaslist)
            if stripit:
                value = map(str.strip,value)
            if mapfunc:
                value = map(mapfunc,value)
        return value


    @classmethod
    def config(cls, settings, prefix=None, defaultval=None,**kwargs):
        
        if not prefix:
            prefix = cls.__name__.lower()

        if cls.myconf:
            cls.__getMyConf(settings ,prefix,defaultval  )
        else:
            cls.__getAllConf(settings ,prefix,defaultval  )

        cls.confended()


    @classmethod
    def conf2dict(cls):
        valdict = {}
        for k in cls.myconf:
            if isinstance(k, string):
                prop = k
            else:
                prop = k[0]
            try:
                valdict[prop] = getattr(cls,prop)
            except Exception as e:
                l.exception(f"{prop} configured as a key, but no value found")
        return valdict




from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine
import json

import logging
l = logging.getLogger(__name__)

from .jsonutils import jsonwalker


class CASTTABLE:
    __code2type = {0: str, 1: int, 2: float}

    __type2code = {
        int: 1,
        float: 2,
    }

    @staticmethod
    def gettype(code):
        return CASTTABLE.__code2type.get(int(code), str)

    @staticmethod
    def getcode(datatype):
        code = CASTTABLE.__type2code.get(datatype, None)
        if code is None:
            return ""
        return str(code).rjust(2, "0")

    # def getencoded(value,encrypted):
    #     code = CASTTABLE.getcode(value)
    #     if code:
    #         return "$".join([code,encrypted])

    # def getdecoded(value):
    #     parts = "$".split(value)
    #     if len(parts)==2:
    #         casttype = CASTTABLE.get(int(parts[0] ),str)
    #     decrypted_value = casttype(self.engine.decrypt(parts[-1]))
    #     return decrypted_value


class CryptUtil:
    def __init__(
        self,
        key=None,
        engine=None,
        padding=None,
    ) -> None:
        self._key = key
        if not engine:
            engine = AesEngine
        self.engine = engine()
        if isinstance(self.engine, AesEngine):
            self.engine._set_padding_mechanism(padding)

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._key = value

    def _update_key(self):
        key = self._key() if callable(self._key) else self._key
        self.engine._update_key(key)

    def encryptval(self, value: any,preformat = None,postformat = None) -> str:
        """
        try to encrypt the value, regardless of what it is. Adds a CODE$ at the beginning if a number type is recognised to allow cast back
        """
        try:
            self._update_key()
            v = self.engine.encrypt(value if preformat is None else preformat(value))
            if code := CASTTABLE.getcode(type(value)):
                v = code + "$" + v
            return v if postformat is None else postformat(v)

        except Exception as e:
            l.exception(f"******************exception while processing {value}:\n{e}")
            return ''

    def decryptval(self, value) -> any:
        """
        Stringify value, try to get 'CODE$' part, decrypt the rest of the string and cast it according to CODE, if present
        """
        self._update_key()
        parts = str(value).split(
            "$"
        )  # value should already be a str, but plain uncrypted object may be other things
        casttype = None
        if len(parts) == 2:
            casttype = CASTTABLE.gettype(int(parts[0]))
        try:
            decrypted_value = self.engine.decrypt(parts[-1])
        except Exception as e:
            # can't decrypt, it was a old plain value
            l.debug("can't decrypt, sending back the plain value")
            return value

        if casttype:
            decrypted_value = casttype(decrypted_value)
        return decrypted_value

    def encryptRecursive(self, value: any, preformat = None, postformat = None, skip_keys:list = [] ) -> str:
        """
        Recursively walk through value and crypt elemets separately and json dump it to string
        """        
        self._update_key()        
        w = jsonwalker(lambda val: self.encryptval(val,preformat,postformat), skip_keys)
        newstruct = json.dumps(w.walk(value))
        #l.debug(f"encrypted in {newstruct}")
        return newstruct

    def decryptRecursive(self, value, skip_keys:list = []):
        """
        JSON parse Value and recursively walk through it and decrypt elemets separately
        """ 
        w = jsonwalker(self.decryptval, skip_keys)
        try:
            val = json.loads(value)
        except Exception as e:
            val = {}
            if value:
                l.info(f"exception decrypting {value}:{e}")
        newstruct = w.walk(val)        
        return newstruct


#!/usr/bin/env python

#from distutils.core import setup

from setuptools import setup,find_packages

import os
here = os.path.abspath(os.path.dirname(__file__))
readme = open(os.path.join(here, 'README.md'), 'r').read()
changelog = open(os.path.join(here, 'change.log'), 'r').read()


setup(name='ppss_pyramidutils',
      version='1.6.1.1',
      description='Simple utils to handle data from ini files in Pyramid for python 2.7 & 3',
      long_description=readme + "\n\n\n" + changelog,
      long_description_content_type="text/markdown",
      author='pdepmcp',
      author_email='pdepmcp@gmail.com',
      install_requires=[
        'six',
        'sqlalchemy_utils',
        'cryptography'
      ],
      keywords="pyramid module utils accelerator",
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        "Operating System :: OS Independent",
        'Framework :: Pyramid',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
      ],
      packages=find_packages()
     )



from contextlib import contextmanager

import time
import logging
l = logging.getLogger(__name__)

################# create session factory from pyramid settings
from sqlalchemy import engine_from_config
from sqlalchemy.orm import sessionmaker

def engineFromSettings(config):
  engine = engine_from_config(config, "sqlalchemy.")
  return engine

def factoryFormSettings(config):
  session_factory = None
  try:
    engine = engineFromSettings(config)
    session_factory = sessionmaker(bind=engine)
  except Exception as e:
    l.exception("can't create session factory")
    time.sleep(10)
  return session_factory


#### allow usage of "with" clause with session
@contextmanager
def session_scope(sessionFactory):
    session = sessionFactory()
    l.debug ("new session created")
    try:
        yield session
        l.debug ("committing")
        session.commit()
    except Exception as e:
        session.rollback()
        l.exception("rollback for exception {}".format(e))
        raise e
    finally:
        session.close()
        l.debug ("session closed")